local _G						= getfenv(0)

local ipairs					= _G.ipairs
local next						= _G.next
local select					= _G.select
local unpack					= _G.unpack

local table_concat				= _G.table.concat
local table_insert				= _G.table.insert
local table_remove				= _G.table.remove
local table_sort				= _G.table.sort
local table_wipe				= _G.table.wipe

local CanViewOfficerNote		= _G.CanViewOfficerNote
local GetGuildRosterInfo		= _G.GetGuildRosterInfo
local GetGuildRosterMOTD		= _G.GetGuildRosterMOTD
local GetNumGuildMembers		= _G.GetNumGuildMembers
local GetQuestDifficultyColor	= _G.GetQuestDifficultyColor
local GuildRoster				= _G.GuildRoster
local IsInGuild					= _G.IsInGuild
local UnitInGroup				= _G.UnitInGroup

local ElkGuild = _G.LibStub("AceAddon-3.0"):NewAddon("ElkGuild", "AceBucket-3.0", "AceConsole-3.0", "AceEvent-3.0")

local LDB = _G.LibStub("LibDataBroker-1.1")
local LDBIcon = _G.LibStub("LibDBIcon-1.0")
local LibQTip = _G.LibStub("LibQTip-1.0")

local DBdefaults = {
	profile = {
		minimap = {
			hide = false,
		},
		sorting = "ZONE:LEVEL:NAME:CLASS:RANK:NOTE:OFFICERNOTE",
		filterClass = {
			DEATHKNIGHT = true,
			DRUID = true,
			HUNTER = true,
			MAGE = true,
			MONK = true,
			PALADIN = true,
			PRIEST = true,
			ROGUE = true,
			SHAMAN = true,
			WARLOCK = true,
			WARRIOR = true,
		},
	},
}


local L = _G.setmetatable({
	TEXT_NOGUILD = "No Guild",
	TEXT_UPDATING = "Updating...",
	TOOLTIP_COLUMN_HEADER_NAME = "Name",
	TOOLTIP_COLUMN_HEADER_LEVEL = "Level",
	TOOLTIP_COLUMN_HEADER_ZONE = "Zone",
	TOOLTIP_COLUMN_HEADER_NOTES = "Notes",
	TOOLTIP_COLUMN_HEADER_RANK = "Rank",
	TOOLTIP_TEXT_MOTD = "MOTD:",
	TOOLTIP_TEXT_NOGUILD = "You aren't in a guild.",
}, {__index = function(self, key) return key end })

-- -----
-- Roster Updater
-- -----
local lastGuildRoster = 0
local updateFrame = _G.CreateFrame("frame")
updateFrame:Hide()
updateFrame:SetScript("OnUpdate", function(self, elapsed)
	if (GetTime() < nextGuildRoster) then return end
	GuildRoster()
end)

_G.hooksecurefunc("GuildRoster", function()
	nextGuildRoster = GetTime() + 15
end)

-- -----
-- player table cache
-- -----
local tablecache = {}
local function recyclePlayerTable(t)
	while #t > 0 do
		table_insert(tablecache, table_remove(t))
	end
	return t
end
local function getPlayerTable(...)
	local t
	if #tablecache > 0 then
		t = table_remove(tablecache)
	else
		t = {}
	end
	for i = 1, select('#', ...) do
		t[i] = select(i, ...)
	end
	return t
end

-- -----
local function do_OnEnter(frame)
	local tooltip = LibQTip:Acquire("ElkGuildTip")
	tooltip:SmartAnchorTo(frame)
	tooltip:SetAutoHideDelay(0.1, frame)
	tooltip:EnableMouse(true)
	if IsInGuild() then
		ElkGuild:UpdateTooltip()
		tooltip:Show()
		updateFrame:Show()
	else
		ElkGuild:UpdateTooltip_noGuild()
		tooltip:Show()
	end
end

local function do_OnLeave()
	updateFrame:Hide()
end

local function do_OnClick(frame, button)
	if button == "RightButton" then
		_G.InterfaceOptionsFrame_OpenToCategory("ElkGuild")
	else
		_G.ToggleGuildFrame()
	end
end

function ElkGuild:OnInitialize()
	self.players = {}

	self.dbo = _G.LibStub("LibDataBroker-1.1"):NewDataObject("ElkGuild", {
			type = "data source",
			text = L["TEXT_UPDATING"],
			icon = [[Interface\Addons\ElkGuild\ElkGuild]],
			OnEnter = do_OnEnter,
			OnLeave = do_OnLeave,
			OnClick = do_OnClick,
		})

	self.db = _G.LibStub("AceDB-3.0"):New("ElkGuildDB", DBdefaults, true)
	self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig")
	self:LoadSort()

	LDBIcon:Register("ElkGuild", self.dbo, self.db.profile.minimap)
end

function ElkGuild:OnEnable()
	self:RegisterEvent("PLAYER_GUILD_UPDATE")
	self:PLAYER_GUILD_UPDATE()
end

function ElkGuild:RefreshConfig()
	-- minimap icon
	LDBIcon:Refresh("ElkGuild", self.db.profile.minimap)
	
	self:LoadSort()
	self:UpdateData()
	self:UpdateText()
	self:UpdateTooltip()
end

function ElkGuild:PLAYER_GUILD_UPDATE(unit)
	if unit and unit ~= "player" then return end
--	self:print("Debug: PLAYER_GUILD_UPDATE fired...")
	if IsInGuild() then
		if not self.bucket_GUILD_ROSTER_UPDATE then
			self.bucket_GUILD_ROSTER_UPDATE = self:RegisterBucketEvent("GUILD_ROSTER_UPDATE", .5)
			self:RegisterEvent("GUILD_MOTD")
		end
		GuildRoster()
	else
		if self.bucket_GUILD_ROSTER_UPDATE then
			self:UnregisterBucket(self.bucket_GUILD_ROSTER_UPDATE)
			self.bucket_GUILD_ROSTER_UPDATE = nil
			self:UnregisterEvent("GUILD_MOTD")
		end
		self:UpdateText()
	end
end

function ElkGuild:GUILD_ROSTER_UPDATE()
--~ 	self:Print("GUILD_ROSTER_UPDATE fired...")
	nextGuildRoster = GetTime() + 15

	self:UpdateData()

	-- update the LDB text
	self:UpdateText()
	
	-- update the tooltip
	self:UpdateTooltip()
end

function ElkGuild:GUILD_MOTD()
	self:UpdateTooltip()
end

-- -----
-- FILTERING
-- -----
function ElkGuild:checkFilter(classFileName, level, zone)
	return true
end

-- -----
-- SORTING
-- -----
local SORT_CONST = {
	[1] = "NAME",			["NAME"] = 1,			[-1] = "-NAME",			["-NAME"] = -1,
	[3] = "RANK",			["RANK"] = 3,			[-3] = "-RANK",			["-RANK"] = -3,
	[4] = "LEVEL",			["LEVEL"] = 4,			[-4] = "-LEVEL",		["-LEVEL"] = -4,
	[5] = "CLASS",			["CLASS"] = 5,			[-5] = "-CLASS",		["-CLASS"] = -5,
	[6] = "ZONE",			["ZONE"] = 6,			[-6] = "-ZONE",			["-ZONE"] = -6,
	[7] = "NOTE",			["NOTE"] = 7,			[-7] = "-NOTE",			["-NOTE"] = -7,
	[8] = "OFFICERNOTE",	["OFFICERNOTE"] = 8,	[-8] = "-OFFICERNOTE",	["-OFFICERNOTE"] = -8,
}

local sortOrder = {SORT_CONST.ZONE, SORT_CONST.LEVEL, SORT_CONST.NAME, SORT_CONST.CLASS, SORT_CONST.RANK, SORT_CONST.NOTE, SORT_CONST.OFFICERNOTE}
local sortIndex = 1
local function sorter(a, b)
	local sortField = sortOrder[sortIndex] or 1 -- SORT_CONST.NAME since name is unique
	local asc
	if (sortField > 0) then
		asc = true
	else
		sortField = -sortField
		asc = false
	end
	local aVal = a[sortField]
	local bVal = b[sortField]
	if (bVal and (not aVal or aVal < bVal)) then
		sortIndex = 1
		return asc
	elseif (aVal and (not bVal or aVal > bVal)) then
		sortIndex = 1
		return not asc
	else
		sortIndex = sortIndex + 1
		return sorter(a, b)
	end
end

function ElkGuild:LoadSort()
	local sortString = self.db.profile.sorting
	table_wipe(sortOrder)
	for key in sortString:gmatch("(%-?[A-Z]+)") do
		local sortField = SORT_CONST[key]
		if (sortField) then
			table_insert(sortOrder, sortField)
		end
	end
end

function ElkGuild:SaveSort()
	local sortString = {}
	for _, v in ipairs(sortOrder) do
		table_insert(sortString, SORT_CONST[v])
	end
	self.db.profile.sorting = table_concat(sortString, ":")
end

function ElkGuild:UpdateData()
	if IsInGuild() then
		local players = recyclePlayerTable(self.players)
		local playersShown = 0
		local playersOnline = 0
		local numGuildMembers = GetNumGuildMembers()
		local name, rank, rankIndex, level, class, zone, note, officernote, online, isAway, classFileName, achievementPoints, achievementRank, isMobile
		for i = 1, numGuildMembers, 1 do
			name, rank, rankIndex, level, class, zone, note, officernote, online, isAway, classFileName, achievementPoints, achievementRank, isMobile = GetGuildRosterInfo(i)
			if (not note or note == "") then note = false end
			if (not officernote or officernote == "") then officernote = false end
			if (online) then
				playersOnline = playersOnline + 1
				if self:checkFilter(classFileName, level, zone) then
					playersShown = playersShown + 1
					table_insert(players, getPlayerTable(
						name or _G.UNKNOWN,
						rank or _G.UNKNOWN,
						rankIndex or -1,
						level or -1,
						class or _G.UNKNOWN,
						isMobile and _G.REMOTE_CHAT or (zone or _G.UNKNOWN),
						note,
						officernote,
						online,
						isAway,
						classFileName,
						achievementPoints,
						achievementRank,
						isMobile
					))
				end
			end
		end
		table_sort(players, sorter)
		
		self.players = players
		self.playersShown = playersShown
		self.playersOnline = playersOnline
		self.playersTotal = GetNumGuildMembers(true)
	else
		self.players = recyclePlayerTable(self.players)
		self.playersShown = 0
		self.playersOnline = 0
		self.playersTotal = 0
	end
end

function ElkGuild:UpdateText()
	if IsInGuild() then
		self.dbo.text = self.playersOnline.."/"..self.playersTotal
	else
		self.dbo.text = L["TEXT_NOGUILD"]
	end
end

-- -----
-- TOOLTIP
-- -----
local function OnHeaderClick(frame, sortString, button)
	local leftSort, rightSort = sortString:match("([A-Z]*):?([A-Z]*)")
	local targetSort = (button == "RightButton") and rightSort or leftSort
	if (targetSort == "") then
		return
	else
		targetSort = SORT_CONST[targetSort]
	end
	local mainSort = sortOrder[1]
	if (mainSort == targetSort) then
		sortOrder[1] = -targetSort
	elseif (mainSort == -targetSort) then
		sortOrder[1] = targetSort
	else
		for k, v in ipairs(sortOrder) do
			if (v == targetSort or v == -targetSort) then
				table_remove(sortOrder, k)
				break
			end
		end
		table_insert(sortOrder, 1, targetSort)
	end

	ElkGuild:SaveSort()
	table_sort(ElkGuild.players, sorter)
	ElkGuild:UpdateTooltip()
end

local function OnNameClick(frame, name, button)
	if not name then return end
	if _G.IsAltKeyDown() then
		_G.InviteUnit(name)
	else
		_G.SetItemRef("player:"..name, "|Hplayer:"..name.."|h["..name.."|h", "LeftButton")
	end
end

local function RGB2HEX(red, green, blue)
	return ("|cff%02x%02x%02x"):format(red * 255, green * 255, blue * 255)
end

function ElkGuild:UpdateTooltip_noGuild()
	if not LibQTip:IsAcquired("ElkGuildTip") then
		return
	end
	local tooltip = LibQTip:Acquire("ElkGuildTip")
	tooltip:Clear()
	tooltip:SetColumnLayout(1, "CENTER")
	tooltip:AddHeader("|cfffed100Elk|cffffffffGuild")
	tooltip:AddLine(L["TOOLTIP_TEXT_NOGUILD"])
end

local statusMap = {
	[0] = "",
	[1] = "|TInterface\\FriendsFrame\\StatusIcon-Away:14:14:0:0:16:16:0:16:0:16|t",
	[2] = "|TInterface\\FriendsFrame\\StatusIcon-DnD:14:14:0:0:16:16:0:16:0:16|t",
	[3] = "|TInterface\\ChatFrame\\UI-ChatIcon-ArmoryChat:14:14:0:0:16:16:0:16:0:16:73:177:73|t",
	[4] = "|TInterface\\ChatFrame\\UI-ChatIcon-ArmoryChat-AwayMobile:14:14:0:0:16:16:0:16:0:16|t",
	[5] = "|TInterface\\ChatFrame\\UI-ChatIcon-ArmoryChat-BusyMobile:14:14:0:0:16:16:0:16:0:16|t",
}
function getSortArrow(sorting)
	if (sortOrder[1] == SORT_CONST[sorting]) then
		return " |TInterface\\Buttons\\UI-SortArrow:8:8:0:-1:16:8:0:8:0:8|t"
	elseif (sortOrder[1] == SORT_CONST["-"..sorting]) then
		return " |TInterface\\Buttons\\UI-SortArrow:8:8:0:0:16:8:0:8:8:0|t"
	elseif (sortOrder[2] == SORT_CONST[sorting]) then
		return " |TInterface\\Buttons\\UI-SortArrow:8:8:0:-1:16:8:0:8:0:8:127:127:127|t"
	elseif (sortOrder[2] == SORT_CONST["-"..sorting]) then
		return " |TInterface\\Buttons\\UI-SortArrow:8:8:0:0:16:8:0:8:8:0:127:127:127|t"
	else
		return ""
	end
end

function ElkGuild:UpdateTooltip()
	if not LibQTip:IsAcquired("ElkGuildTip") then
		updateFrame:Hide()
		return
	end
	local tooltip = LibQTip:Acquire("ElkGuildTip")
	tooltip:Clear()

	local lineNum
	tooltip:SetColumnLayout(6, "CENTER", "LEFT", "CENTER", "CENTER", "CENTER", "RIGHT")

	lineNum = tooltip:AddHeader()
	tooltip:SetCell(lineNum, 1, "|cfffed100Elk|cffffffffGuild", tooltip:GetHeaderFont(), "CENTER", 6)

	lineNum = tooltip:AddHeader()
	tooltip:SetCell(lineNum, 1, "|cffffd200"..L["TOOLTIP_TEXT_MOTD"], nil, "LEFT", 6)
	lineNum = tooltip:AddLine()
	tooltip:SetCell(lineNum, 1, (GetGuildRosterMOTD()), nil, "LEFT", 6, nil, nil, nil, 0)
	lineNum = tooltip:AddLine(" ")

	lineNum = tooltip:AddLine()
	tooltip:SetCell(lineNum, 2, "|cffffd200"..getSortArrow("CLASS")..L["TOOLTIP_COLUMN_HEADER_NAME"]..getSortArrow("NAME"), nil, "CENTER")
	tooltip:SetCellScript(lineNum, 2, "OnMouseUp", OnHeaderClick, "NAME:CLASS")
	tooltip:SetCell(lineNum, 3, "|cffffd200"..L["TOOLTIP_COLUMN_HEADER_LEVEL"]..getSortArrow("LEVEL"), nil, "CENTER")
	tooltip:SetCellScript(lineNum, 3, "OnMouseUp", OnHeaderClick, "LEVEL")
	tooltip:SetCell(lineNum, 4, "|cffffd200"..L["TOOLTIP_COLUMN_HEADER_ZONE"]..getSortArrow("ZONE"), nil, "CENTER")
	tooltip:SetCellScript(lineNum, 4, "OnMouseUp", OnHeaderClick, "ZONE")
	tooltip:SetCell(lineNum, 5, "|cffffd200"..getSortArrow("OFFICERNOTE")..L["TOOLTIP_COLUMN_HEADER_NOTES"]..getSortArrow("NOTE"), nil, "CENTER")
	tooltip:SetCellScript(lineNum, 5, "OnMouseUp", OnHeaderClick, "NOTE:OFFICERNOTE")
	tooltip:SetCell(lineNum, 6, "|cffffd200"..L["TOOLTIP_COLUMN_HEADER_RANK"]..getSortArrow("RANK"), nil, "CENTER")
	tooltip:SetCellScript(lineNum, 6, "OnMouseUp", OnHeaderClick, "RANK")
	lineNum = tooltip:AddSeparator(2)
	
	local canViewOfficerNote = CanViewOfficerNote()
	for _, player in ipairs(self.players) do
		local name, rank, rankIndex, level, class, zone, note, officernote, online, isAway, classFileName, achievementPoints, achievementRank, isMobile = unpack(player)
		local sameGroup = UnitInParty(name) or UnitInRaid(name)
		local status = statusMap[isAway + (isMobile and 3 or 0)]
		local levelcolor = GetQuestDifficultyColor(level)
		levelcolor = RGB2HEX(levelcolor.r, levelcolor.g, levelcolor.b)
		local tipnote = note and "["..note.."]" or "-"
		if canViewOfficerNote then
			tipnote = tipnote .. (officernote and " ["..officernote.."]" or " -")
		end
		lineNum = tooltip:AddLine()
		tooltip:SetCell(lineNum, 1, sameGroup and "|TInterface\\Buttons\\UI-CheckBox-Check:18|t" or "")		-- same group?
		tooltip:SetCell(lineNum, 2, status.."|c".._G.RAID_CLASS_COLORS[classFileName].colorStr..name)		-- status / name
		tooltip:SetCell(lineNum, 3, levelcolor..level)														-- level
		tooltip:SetCell(lineNum, 4, zone)																	-- zone
		tooltip:SetCell(lineNum, 5, tipnote)																-- notes
		tooltip:SetCell(lineNum, 6, rank)																	-- rank
		tooltip:SetLineScript(lineNum, "OnMouseUp", OnNameClick, name)
	end
	tooltip:UpdateScrolling()
end